import pytest
from pathlib import Path
from haystack import Pipeline
from haystack.core.errors import PipelineError

from hayhooks.server.pipelines.registry import _PipelineRegistry


@pytest.fixture
def pipeline_registry():
    return _PipelineRegistry()


@pytest.fixture
def sample_pipeline_yaml():
    return (Path(__file__).parent / "test_files" / "working_pipelines" / "basic_rag_pipeline.yml").read_text()


def test_add_pipeline(pipeline_registry, sample_pipeline_yaml):
    result = pipeline_registry.add("test_pipeline", sample_pipeline_yaml)

    expected_pipeline = Pipeline.loads(sample_pipeline_yaml)
    assert result == expected_pipeline
    assert pipeline_registry.get("test_pipeline") == expected_pipeline


def test_add_duplicate_pipeline(pipeline_registry, sample_pipeline_yaml):
    result = pipeline_registry.add("test_pipeline", sample_pipeline_yaml)

    expected_pipeline = Pipeline.loads(sample_pipeline_yaml)
    assert result == expected_pipeline

    with pytest.raises(ValueError, match="A pipeline with name test_pipeline is already in the registry"):
        pipeline_registry.add("test_pipeline", sample_pipeline_yaml)


def test_add_invalid_pipeline(pipeline_registry, mocker):
    mocker.patch('haystack.Pipeline.loads', side_effect=PipelineError("Invalid pipeline"))

    with pytest.raises(ValueError, match="Unable to parse Haystack Pipeline test_pipeline"):
        pipeline_registry.add("test_pipeline", "invalid yaml")


def test_remove_pipeline(pipeline_registry, sample_pipeline_yaml):
    pipeline_registry.add("test_pipeline", sample_pipeline_yaml)
    pipeline_registry.remove("test_pipeline")

    assert pipeline_registry.get("test_pipeline") is None


def test_remove_nonexistent_pipeline(pipeline_registry):
    pipeline_registry.remove("nonexistent_pipeline")


def test_get_nonexistent_pipeline(pipeline_registry):
    assert pipeline_registry.get("nonexistent_pipeline") is None


def test_get_names(pipeline_registry, sample_pipeline_yaml, mocker):
    mocker.patch('haystack.Pipeline.loads', return_value=mocker.Mock(spec=Pipeline))

    pipeline_registry.add("pipeline1", sample_pipeline_yaml)
    pipeline_registry.add("pipeline2", sample_pipeline_yaml)

    names = pipeline_registry.get_names()
    assert sorted(names) == ["pipeline1", "pipeline2"]



from fastapi import HTTPException
from fastapi.concurrency import run_in_threadpool
from fastapi.responses import JSONResponse

from hayhooks.server.pipelines import registry
from hayhooks.server.pipelines.models import (
    PipelineDefinition,
    convert_component_output,
    get_request_model,
    get_response_model,
)


def deploy_pipeline_def(app, pipeline_def: PipelineDefinition):
    try:
        pipe = registry.add(pipeline_def.name, pipeline_def.source_code)
    except ValueError as e:
        raise HTTPException(status_code=409, detail=f"{e}") from e
    except Exception as e:
        raise HTTPException(status_code=500, detail=f"{e}") from e

    PipelineRunRequest = get_request_model(pipeline_def.name, pipe.inputs())
    PipelineRunResponse = get_response_model(pipeline_def.name, pipe.outputs())

    # There's no way in FastAPI to define the type of the request body other than annotating
    # the endpoint handler. We have to ignore the type here to make FastAPI happy while
    # silencing static type checkers (that would have good reasons to trigger!).
    async def pipeline_run(pipeline_run_req: PipelineRunRequest) -> JSONResponse:  # type: ignore
        result = await run_in_threadpool(pipe.run, data=pipeline_run_req.dict())
        final_output = {}
        for component_name, output in result.items():
            final_output[component_name] = convert_component_output(output)

        return JSONResponse(PipelineRunResponse(**final_output).model_dump(), status_code=200)

    app.add_api_route(
        path=f"/{pipeline_def.name}",
        endpoint=pipeline_run,
        methods=["POST"],
        name=pipeline_def.name,
        response_model=PipelineRunResponse,
        tags=["pipelines"],
    )
    app.openapi_schema = None
    app.setup()

    return {"name": pipeline_def.name}



from collections.abc import Callable as CallableABC
from types import GenericAlias
from typing import Callable, Optional, Union, get_args, get_origin
from loguru import logger


def is_callable_type(t):
    """Check if a type is any form of callable"""
    if t in (Callable, CallableABC):
        return True

    # Check origin type
    origin = get_origin(t)
    if origin in (Callable, CallableABC):
        return True

    # Handle Optional/Union types
    if origin in (Union, type(Optional[int])):  # type(Optional[int]) handles runtime Optional type
        args = get_args(t)
        return any(is_callable_type(arg) for arg in args)

    return False


def handle_unsupported_types(
    type_: type, types_mapping: dict, skip_callables: bool = True
) -> Union[GenericAlias, type, None]:
    logger.debug(f"Handling unsupported type: {type_}")

    if skip_callables and is_callable_type(type_):
        logger.warning(f"Skipping callable type: {type_}")
        return None

    # Handle generic types (like List, Optional, etc.)
    origin = get_origin(type_)
    if origin is not None:
        args = get_args(type_)
        # Map the inner types using the same mapping
        mapped_args = tuple(handle_unsupported_types(arg, types_mapping, skip_callables) or arg for arg in args)
        # Reconstruct the generic type with mapped arguments
        return origin[mapped_args]

    if type_ in types_mapping:
        logger.debug(f"Mapping type: {type_} to {types_mapping[type_]}")
        return types_mapping[type_]

    logger.debug(f"Returning original type: {type_}")
    return type_



from hayhooks.server.handlers import deploy
from hayhooks.server.handlers import draw
from hayhooks.server.handlers import status
from hayhooks.server.handlers import undeploy


__all__ = ["deploy", "draw", "status", "undeploy"]



from typing import cast

from fastapi import HTTPException
from fastapi.routing import APIRoute
from hayhooks.server import app
from hayhooks.server.pipelines import registry


@app.post("/undeploy/{pipeline_name}", tags=["config"])
async def deploy(pipeline_name: str):
    if pipeline_name not in registry.get_names():
        raise HTTPException(status_code=404)

    new_routes = []
    for route in app.router.routes:
        route = cast(APIRoute, route)
        if route.name != pipeline_name:
            new_routes.append(route)

    app.router.routes = new_routes
    app.openapi_schema = None
    app.setup()
    registry.remove(pipeline_name)



from pathlib import Path
from urllib.parse import urljoin

import click
import requests


@click.command()
@click.pass_obj
@click.option('-n', '--name')
@click.argument('pipeline_file', type=click.File('r'))
def deploy(server_conf, name, pipeline_file):
    server, disable_ssl = server_conf
    if name is None:
        name = Path(pipeline_file.name).stem
    resp = requests.post(
        urljoin(server, "deploy"),
        json={"name": name, "source_code": str(pipeline_file.read())}, 
        verify=not disable_ssl
    )

    if resp.status_code >= 400:
        click.echo(f"Error deploying pipeline: {resp.json().get('detail')}")
    else:
        click.echo(f"Pipeline successfully deployed with name: {resp.json().get('name')}")



from pathlib import Path
from urllib.parse import urljoin

import click
import requests
from requests import ConnectionError


@click.command()
@click.pass_obj
@click.argument('pipeline_name')
def undeploy(server_conf, pipeline_name):
    server, disable_ssl = server_conf
    try:
        resp = requests.post(urljoin(server, f"undeploy/{pipeline_name}"), verify=not disable_ssl)

        if resp.status_code >= 400:
            click.echo(f"Cannot undeploy pipeline: {resp.json().get('detail')}")
        else:
            click.echo(f"Pipeline successfully undeployed")
    except ConnectionError:
        click.echo("Hayhooks server is not responding. To start one, run `hayooks run`")



from pandas import DataFrame
from pydantic import BaseModel, ConfigDict, create_model
from hayhooks.server.utils.create_valid_type import handle_unsupported_types
from haystack import Document


class PipelineDefinition(BaseModel):
    name: str
    source_code: str


DEFAULT_TYPES_MAPPING = {
    DataFrame: dict,
    Document: dict,
}


def get_request_model(pipeline_name: str, pipeline_inputs):
    """
    Inputs have this form:
    {
        'first_addition': { <-- Component Name
            'value': {'type': <class 'int'>, 'is_mandatory': True}, <-- Input
            'add': {'type': typing.Optional[int], 'is_mandatory': False, 'default_value': None}, <-- Input
        },
        'second_addition': {'add': {'type': typing.Optional[int], 'is_mandatory': False}},
    }
    """
    request_model = {}
    config = ConfigDict(arbitrary_types_allowed=True)

    for component_name, inputs in pipeline_inputs.items():
        component_model = {}
        for name, typedef in inputs.items():
            try:
                input_type = handle_unsupported_types(type_=typedef["type"], types_mapping=DEFAULT_TYPES_MAPPING)
            except TypeError as e:
                print(f"ERROR at {component_name!r}, {name}: {typedef}")
                raise e

            if input_type is not None:
                component_model[name] = (
                    input_type,
                    typedef.get("default_value", ...),
                )
        request_model[component_name] = (create_model("ComponentParams", **component_model, __config__=config), ...)

    return create_model(f"{pipeline_name.capitalize()}RunRequest", **request_model, __config__=config)


def get_response_model(pipeline_name: str, pipeline_outputs):
    """
    Outputs have this form:
    {
        'second_addition': { <-- Component Name
            'result': {'type': "<class 'int'>"}  <-- Output
        },
    }
    """
    response_model = {}
    config = ConfigDict(arbitrary_types_allowed=True)
    for component_name, outputs in pipeline_outputs.items():
        component_model = {}
        for name, typedef in outputs.items():
            output_type = typedef["type"]
            component_model[name] = (
                handle_unsupported_types(type_=output_type, types_mapping=DEFAULT_TYPES_MAPPING),
                ...,
            )
        response_model[component_name] = (create_model("ComponentParams", **component_model, __config__=config), ...)

    return create_model(f"{pipeline_name.capitalize()}RunResponse", **response_model, __config__=config)


def convert_value_to_dict(value):
    """Convert a single value to a dictionary if possible"""
    if hasattr(value, "to_dict"):
        if "init_parameters" in value.to_dict():
            return value.to_dict()["init_parameters"]
        return value.to_dict()
    elif hasattr(value, "model_dump"):
        return value.model_dump()
    elif isinstance(value, dict):
        return {k: convert_value_to_dict(v) for k, v in value.items()}
    elif isinstance(value, list):
        return [convert_value_to_dict(item) for item in value]
    else:
        return value


def convert_component_output(component_output):
    """
    Converts component outputs to dictionaries that can be validated against response model.
    Handles nested structures recursively.

    Args:
        component_output: Dict with component outputs

    Returns:
        Dict with all nested objects converted to dictionaries
    """
    if isinstance(component_output, dict):
        return {name: convert_value_to_dict(data) for name, data in component_output.items()}

    return convert_value_to_dict(component_output)



from fastapi import HTTPException
from hayhooks.server import app
from hayhooks.server.pipelines import registry


@app.get("/status")
async def status():
    pipelines = registry.get_names()
    return {"status": "Up!", "pipelines": pipelines}



from haystack.core.component import component


@component
class Hello:
    @component.output_types(output=str)
    def run(self, word: str = "world"):
        """
        Takes a string in input and returns "Hello, <string>!"
        in output.
        """
        return {"output": f"Hello, {word}!"}



from fastapi.testclient import TestClient
from hayhooks.server import app
from pathlib import Path

client = TestClient(app)


def test_gracefully_handle_deploy_exception():
    pipeline_name = "broken_rag_pipeline"
    pipeline_def = (Path(__file__).parent / "test_files" / "broken_rag_pipeline.yml").read_text()

    deploy_response = client.post("/deploy", json={"name": pipeline_name, "source_code": pipeline_def})
    assert deploy_response.status_code == 500
    assert "Couldn't deserialize component 'llm'" in deploy_response.json()["detail"]



from urllib.parse import urljoin

import click
import requests
from requests.exceptions import ConnectionError


@click.command()
@click.pass_obj
def status(server_conf):
    server, disable_ssl = server_conf
    try:
        r = requests.get(urljoin(server, "status"), verify=not disable_ssl)
    except ConnectionError:
        click.echo("Hayhooks server is not responding. To start one, run `hayhooks run`")
        return

    if r.status_code >= 400:
        body = r.json()
        click.echo(f"Hayhooks server is unhealty: [{r.status_code}] {r. json().get('detail')}")
        return

    click.echo("Hayhooks server is up and running.")
    body = r.json()
    if pipes := body.get("pipelines"):
        click.echo("\nPipelines deployed:")
        for p in pipes:
            click.echo(f"- {p}")



from hayhooks.server.pipelines.models import convert_component_output
from openai.types.completion_usage import CompletionTokensDetails, PromptTokensDetails


def test_convert_component_output_with_nested_models():
    sample_response = [
        {
            'model': 'gpt-4o-mini-2024-07-18',
            'index': 0,
            'finish_reason': 'stop',
            'usage': {
                'completion_tokens': 52,
                'prompt_tokens': 29,
                'total_tokens': 81,
                'completion_tokens_details': CompletionTokensDetails(
                    accepted_prediction_tokens=0, audio_tokens=0, reasoning_tokens=0, rejected_prediction_tokens=0
                ),
                'prompt_tokens_details': PromptTokensDetails(audio_tokens=0, cached_tokens=0),
            },
        }
    ]

    converted_output = convert_component_output(sample_response)

    assert converted_output == [
        {
            'model': 'gpt-4o-mini-2024-07-18',
            'index': 0,
            'finish_reason': 'stop',
            'usage': {
                'completion_tokens': 52,
                'prompt_tokens': 29,
                'total_tokens': 81,
                'completion_tokens_details': {
                    'accepted_prediction_tokens': 0,
                    'audio_tokens': 0,
                    'reasoning_tokens': 0,
                    'rejected_prediction_tokens': 0,
                },
                'prompt_tokens_details': {'audio_tokens': 0, 'cached_tokens': 0},
            },
        }
    ]



import pytest
from fastapi.testclient import TestClient
from hayhooks.server import app
from pathlib import Path

client = TestClient(app)

# Load pipeline definitions from test_files
test_files = Path(__file__).parent / "test_files" / "working_pipelines"
pipeline_data = [{"name": file.stem, "source_code": file.read_text()} for file in test_files.glob("*.yml")]


@pytest.mark.parametrize("pipeline_data", pipeline_data)
def test_deploy_pipeline_def(pipeline_data: dict):
    deploy_response = client.post(
        "/deploy", json={"name": pipeline_data["name"], "source_code": pipeline_data["source_code"]}
    )
    assert deploy_response.status_code == 200

    status_response = client.get("/status")
    assert pipeline_data["name"] in status_response.json()["pipelines"]

    docs_response = client.get("/docs")
    assert docs_response.status_code == 200



from typing import Optional, List
from hayhooks.server.utils.create_valid_type import handle_unsupported_types


def test_handle_simple_type():
    result = handle_unsupported_types(int, {})
    assert result == int


def test_handle_generic_type():
    result = handle_unsupported_types(List[int], {})
    assert result == list[int]


def test_handle_recursive_type():
    class Node:
        def __init__(self, value: int, next: Optional['Node'] = None):
            self.value = value
            self.next = next

    result = handle_unsupported_types(Node, {})
    assert result == Node


def test_handle_circular_reference():
    class A:
        def __init__(self, b: 'B'):
            self.b = b

    class B:
        def __init__(self, a: 'A'):
            self.a = a

    result = handle_unsupported_types(A, {})
    assert result == A  # Adjust assertion based on expected behavior


def test_handle_nested_generics():
    nested_type = dict[str, list[Optional[int]]]
    result = handle_unsupported_types(nested_type, {})
    assert result == nested_type



from collections.abc import Callable as CallableABC
from typing import Any, Callable, Optional, Union

import haystack
import pytest

from hayhooks.server.pipelines.models import get_request_model
from hayhooks.server.utils.create_valid_type import is_callable_type


@pytest.mark.parametrize(
    "t, expected",
    [
        (Callable, True),
        (CallableABC, True),
        (Callable[[int], str], True),
        (Callable[..., Any], True),
        (int, False),
        (str, False),
        (Any, False),
        (Union[int, str], False),
        (Optional[Callable[[haystack.dataclasses.streaming_chunk.StreamingChunk], type(None)]], True),
    ],
)
def test_is_callable_type(t, expected):
    assert is_callable_type(t) == expected


def test_skip_callables_when_creating_pipeline_models():
    pipeline_name = "test_pipeline"
    pipeline_inputs = {
        "generator": {
            "system_prompt": {"type": Optional[str], "is_mandatory": False, "default_value": None},
            "streaming_callback": {
                "type": Optional[Callable[[haystack.dataclasses.streaming_chunk.StreamingChunk], type(None)]],
                "is_mandatory": False,
                "default_value": None,
            },
            "generation_kwargs": {
                "type": Optional[dict[str, Any]],
                "is_mandatory": False,
                "default_value": None,
            },
        }
    }

    request_model = get_request_model(pipeline_name, pipeline_inputs)

    # This line used to throw an error because the Callable type was not handled correctly
    # by the handle_unsupported_types function
    assert request_model.model_json_schema() is not None
    assert request_model.__name__ == "Test_pipelineRunRequest"
    assert "streaming_callback" not in request_model.model_json_schema()["$defs"]["ComponentParams"]["properties"]



from fastapi import FastAPI
import os
import glob
from pathlib import Path
from hayhooks.server.utils.deploy_utils import deploy_pipeline_def, PipelineDefinition
import logging

logger = logging.getLogger("uvicorn.info")


def create_app() -> FastAPI:
    if root_path := os.environ.get("HAYHOOKS_ROOT_PATH"):
        app = FastAPI(root_path=root_path)
    else:
        app = FastAPI()

    # Deploy all pipelines in the pipelines directory
    pipelines_dir = os.environ.get("HAYHOOKS_PIPELINES_DIR")
    if pipelines_dir:
        logger.info(f"Pipelines dir set to: {pipelines_dir}")
        for pipeline_file_path in glob.glob(f"{pipelines_dir}/*.y*ml"):
            name = Path(pipeline_file_path).stem
            with open(pipeline_file_path, "r") as pipeline_file:
                source_code = pipeline_file.read()

            pipeline_defintion = PipelineDefinition(name=name, source_code=source_code)
            deployed_pipeline = deploy_pipeline_def(app, pipeline_defintion)
            logger.info(f"Deployed pipeline: {deployed_pipeline['name']}")
    return app


app = create_app()


@app.get("/")
async def root():
    return {
        "swagger_docs" : "http://localhost:1416/docs",
        "deploy_pipeline" : "http://localhost:1416/deploy",
        "draw_pipeline" : "http://localhost:1416/draw/{pipeline_name}",
        "server_status" : "http://localhost:1416/status",
        "undeploy_pipeline" : "http://localhost:1416/undeploy/{pipeline_name}",
    }
